/*************************************************************************\
*                  Copyright (C) Michael Kerrisk, 2010.                   *
*                                                                         *
* This program is free software. You may use, modify, and redistribute it *
* under the terms of the GNU Affero General Public License as published   *
* by the Free Software Foundation, either version 3 or (at your option)   *
* any later version. This program is distributed without any warranty.    *
* See the file COPYING.agpl-v3 for details.                               *
\*************************************************************************/

/* scm_cred_recv.c

   Used in conjunction with scm_cred_send.c to demonstrate passing of
   process credentials via a UNIX domain socket.

   This program receives credentials sent to a UNIX domain socket.

   Usage is as shown in the usageErr() call below.

   Credentials can exchanged over stream or datagram sockets. This program
   uses stream sockets by default; the "-d" command-line option specifies
   that datagram sockets should be used instead.

   Modified for kFreeBSD and Hurd by Svante Signell April-Aug 2013.
   Added -z option, send a zero byte for credentials with Hurd and kFreeBSD
*/
#include "scm_cred.h"

#ifndef SCM_CREDENTIALS
#define SCM_CREDENTIALS SCM_CREDS;
#endif

int
main(int argc, char *argv[])
{
  struct msghdr msgh;
  struct iovec iov;
  int data;

#ifdef __linux__
  int lfd, sfd, optval, opt;
  struct ucred *ucredp, ucred;
  socklen_t len;
#else
  int lfd, sfd, opt;
  struct cmsgcred *ucredp;
#endif
  ssize_t nr;
  Boolean useDatagramSocket;
  Boolean zerobyte;
  union {
    struct cmsghdr cmh;
#ifdef __linux__
    char control[CMSG_SPACE(sizeof(struct ucred))];
    /* Space large enough to hold a ucred structure */
#else
    char control[CMSG_SPACE(sizeof(struct cmsgcred))];
    /* Space large enough to hold a cmdgcred structure */
#endif
  } control_un;
  struct cmsghdr *cmhp;

  /* Parse command-line arguments */

  useDatagramSocket = FALSE;
  zerobyte = FALSE;

  while ((opt = getopt(argc, argv, "dz")) != -1) {
    switch (opt) {
    case 'd':
      useDatagramSocket = TRUE;
      break;

    case 'z':
      zerobyte = TRUE;
      break;

    default:
      usageErr("%s [-d]\n"
	       "        -d    use datagram socket\n"
	       "        -z    zero credentials byte sent\n", argv[0]);
    }
  }

  /* Create socket bound to well-known address */

  if (remove(SOCK_PATH) == -1 && errno != ENOENT)
    errExit("remove-%s", SOCK_PATH);

  if (useDatagramSocket) {
    printf("Receiving via datagram socket\n");
    sfd = unixBind(SOCK_PATH, SOCK_DGRAM);
    if (sfd == -1)
      errExit("unixBind");
  } else {
    printf("Receiving via stream socket\n");
    lfd = unixListen(SOCK_PATH, 5);
    if (lfd == -1)
      errExit("unixListen");

    sfd = accept(lfd, NULL, NULL);
    if (sfd == -1)
      errExit("accept");
  }

  /* We must set the SO_PASSCRED socket option in order to receive
     credentials */
#if defined(__linux__)
  /* SO_PASSCRED not defined on GNU */
  optval = 1;
  if (setsockopt(sfd, SOL_SOCKET, SO_PASSCRED, &optval, sizeof(optval)) == -1)
    errExit("setsockopt");
#endif

  /* Set 'control_un' to describe ancillary data that we want to receive */

#ifdef __linux__
  control_un.cmh.cmsg_len = CMSG_LEN(sizeof(struct ucred));
#else
  control_un.cmh.cmsg_len = CMSG_LEN(sizeof(struct cmsgcred));
#endif
  control_un.cmh.cmsg_level = SOL_SOCKET;
  control_un.cmh.cmsg_type = SCM_CREDENTIALS;

  /* Set 'msgh' fields to describe 'control_un' */

  msgh.msg_control = control_un.control;
  msgh.msg_controllen = sizeof(control_un.control);

  /* Set fields of 'msgh' to point to buffer used to receive (real)
     data read by recvmsg() */

  msgh.msg_iov = &iov;
  msgh.msg_iovlen = 1;
  iov.iov_base = &data;
  if (zerobyte)
    iov.iov_len = sizeof(char);
  else
    iov.iov_len = sizeof(int);

  msgh.msg_name = NULL;               /* We don't need address of peer */
  msgh.msg_namelen = 0;

  /* Receive real plus ancillary data */

  nr = recvmsg(sfd, &msgh, 0);
  printf("recvmsg() returned %ld bytes\n", (long) nr);
  if (nr == -1)
    errExit("recvmsg");
  else if (nr == 0)
    printf ("peer has performed a shutdown\n");
  else if (nr == 1)
    /* Assume a character was sent */
    if (zerobyte)
      printf("Received char data = '0x%0x'\n", (char)data);
    else
      printf("Received integer data as char = '0x%0x'\n", (char)data);
  else if (nr == 4)
    /* Assume an integer was sent */
      printf("Received integer data = %d\n", data);
  else
    errExit("recvmsg: Faulty data");

  /* Extract credentials information from received ancillary data */
  cmhp = CMSG_FIRSTHDR(&msgh);

#if defined (__GNU__) || defined (__FreeBSD_kernel__)
  int i;
  if (cmhp == NULL)
    fatal("bad cmsg header cmhp == NULL");
  if (cmhp->cmsg_len != CMSG_LEN(sizeof(struct cmsgcred)))
    fatal("bad message length cmhp->cmsg_len=%d", cmhp->cmsg_len);
  if (cmhp->cmsg_level != SOL_SOCKET)
    fatal("cmsg_level != SOL_SOCKET");
  if (cmhp->cmsg_type != SCM_CREDS)
    fatal("cmsg_type != SCM_CREDS");
  ucredp = (struct cmsgcred *) CMSG_DATA(cmhp);
  printf("Received credentials pid=%ld, uid=%ld, euid=%ld, gid=%ld\n",
	 (long) ucredp->cmcred_pid, (long) ucredp->cmcred_uid, 
	 (long) ucredp->cmcred_euid, (long) ucredp->cmcred_gid);
  for (i=0; i<ucredp->cmcred_ngroups; i++)
    printf("ucredp->cmcred_groups[%d]=%d\n", i, ucredp->cmcred_groups[i]);
#else
  if (cmhp == NULL || cmhp->cmsg_len != CMSG_LEN(sizeof(struct ucred)))
    fatal("bad cmsg header / message length");
  if (cmhp->cmsg_level != SOL_SOCKET)
    fatal("cmsg_level != SOL_SOCKET");
  if (cmhp->cmsg_type != SCM_CREDENTIALS)
    fatal("cmsg_type != SCM_CREDENTIALS");
  ucredp = (struct ucred *) CMSG_DATA(cmhp);
  printf("Received credentials pid=%ld, uid=%ld, gid=%ld\n",
	 (long) ucredp->pid, (long) ucredp->uid, (long) ucredp->gid);

  /* The Linux-specific, read-only SO_PEERCRED socket option returns
     credential information about the peer, as described in socket(7) */
  len = sizeof(struct ucred);
  if (getsockopt(sfd, SOL_SOCKET, SO_PEERCRED, &ucred, &len) == -1)
    errExit("getsockopt");

  printf("Credentials from SO_PEERCRED: pid=%ld, euid=%ld, egid=%ld\n",
	 (long) ucred.pid, (long) ucred.uid, (long) ucred.gid);
#endif

  exit(EXIT_SUCCESS);
}
