/************************************************************************* \
*                  Copyright (C) Michael Kerrisk, 2010.                   *
*                                                                         *
* This program is free software. You may use, modify, and redistribute it *
* under the terms of the GNU Affero General Public License as published   *
* by the Free Software Foundation, either version 3 or (at your option)   *
* any later version. This program is distributed without any warranty.    *
* See the file COPYING.agpl-v3 for details.                               *
\*************************************************************************/

/* scm_cred_send.c

   Used in conjunction with scm_cred_recv.c to demonstrate passing of
   process credentials via a UNIX domain socket.

   This program sends credentials to a UNIX domain socket.

   Usage is as shown in the usageErr() call below.

   Credentials can exchanged over stream or datagram sockets. This program
   uses stream sockets by default; the "-d" command-line option specifies
   that datagram sockets should be used instead.

   Modified for kFreeBSD and Hurd by Svante Signell April-Aug 2013.
   Added -z option, send a zero byte for credentials with Hurd and kFreeBSD
*/

#include "scm_cred.h"

#ifndef SCM_CREDENTIALS
#define SCM_CREDENTIALS SCM_CREDS;
#endif

/* 
For Linux it does not matter if '\0' or something else is
sent. Credentials passing is controlled by setsockopt option
SO_PASSCRED used in scm_cred_recv.c
 */

int
main(int argc, char *argv[])
{
  struct msghdr msgh;
  struct iovec iov;
  int sfd, opt;
  char zdata = {0};
  int data;
  ssize_t ns;
  Boolean useDatagramSocket, noExplicit, zerobyte;

#ifdef __linux__
  union {
    struct cmsghdr cmh;
    char control[CMSG_SPACE(sizeof(struct ucred))];
    /* Space large enough to hold a ucred structure */
  } control_un;
#else
  union {
    struct cmsghdr cmh;
    char control[CMSG_SPACE(sizeof(struct cmsgcred))];
    /* Space large enough to hold a cmsgcred structure */
  } control_un;
#endif
  struct cmsghdr *cmhp;

  /* Parse command-line arguments */

  useDatagramSocket = FALSE;
  noExplicit = FALSE;
  zerobyte = FALSE;

  while ((opt = getopt(argc, argv, "dnz")) != -1) {
    switch (opt) {
    case 'd':
      useDatagramSocket = TRUE;
      break;

    case 'n':
      noExplicit = TRUE;
      break;

    case 'z':
      zerobyte = TRUE;
      break;

    default:
#ifdef __linux__
      usageErr("%s [-d] [-n] [-z] [data [PID [UID [GID]]]]\n"
	       "        -d    use datagram socket\n"
	       "        -n    don't construct explicit credentials structure\n"
	       "        -z    send zero credentials byte\n", argv[0]);
#else
      usageErr("%s [-d] [-n] [-z] [data [PID [UID [EUID [GID]]]]]\n"
	       "        -d    use datagram socket\n"
	       "        -n    don't construct explicit credentials structure\n"
	       "        -z    send zero credentials byte\n", argv[0]);
#endif
    }
  }

  /* On Linux, we must transmit at least 1 byte of real data in
     order to send ancillary data */

  msgh.msg_iov = &iov;
  msgh.msg_iovlen = 1;

  /* Data is optionally taken from command line */

  if (zerobyte)
    {
      /* Send a zero credentials byte */
      zdata = '\0';
      iov.iov_base = &zdata;
      iov.iov_len = sizeof(char);
      printf("Sending a zero credentials byte to receive the credentials\n");
      printf("Sent data = '0x%0x'\n", zdata);
    }
  else
    {
      /* Send default integer 12345 or taken from the command line */
      data = (argc > optind) ? atoi(argv[optind]) : 12345;
      iov.iov_base = &data;
      iov.iov_len = sizeof(int);
      printf("Sent data = %d\n", data);
    }

  /* Don't need to specify destination address, because we use
     connect() below */
  msgh.msg_name = NULL;
  msgh.msg_namelen = 0;

  if (noExplicit)
    {
    /* Don't construct an explicit credentials structure. (It
       is not necessary to do so, if we just want the receiver to
       receive our real credentials.) */
    printf("Not explicitly sending a credentials structure\n");
    msgh.msg_control = NULL;
    msgh.msg_controllen = 0;
    } 
  else 
    {
    /* Construct an explicit credentials structure */

#ifdef __linux__
      /* NOTE: ucred does not have euid defined in the struct */
      struct ucred *ucp;
#else
      struct cmsgcred *ucp;
#endif
      /* Set 'msgh' fields to describe 'control_un' */
      msgh.msg_control = control_un.control;
      msgh.msg_controllen = sizeof(control_un.control);

      /* Set message header to describe ancillary data that we want to send */
      cmhp = CMSG_FIRSTHDR(&msgh);
#ifdef __linux__
      cmhp->cmsg_len = CMSG_LEN(sizeof(struct ucred));
      cmhp->cmsg_level = SOL_SOCKET;
      cmhp->cmsg_type = SCM_CREDENTIALS;
      ucp = (struct ucred *) CMSG_DATA(cmhp);
    /*
      We could rewrite the preceding as:

      control_un.cmh.cmsg_len = CMSG_LEN(sizeof(struct ucred));
      control_un.cmh.cmsg_level = SOL_SOCKET;
      control_un.cmh.cmsg_type = SCM_CREDENTIALS;
      ucp = (struct ucred *) CMSG_DATA(CMSG_FIRSTHDR(&msgh));
    */
#else
      cmhp->cmsg_len = CMSG_LEN(sizeof(struct cmsgcred));
      cmhp->cmsg_level = SOL_SOCKET;
      cmhp->cmsg_type = SCM_CREDS;
      ucp = (struct cmsgcred *) CMSG_DATA(cmhp);
#endif

    /* Use sender's own PID, real UID, and real GID, unless
       alternate values were supplied on the command line */
#ifdef __linux__
    ucp->pid = (argc > optind + 1 && strcmp(argv[optind + 1], "-") != 0) ?
      atoi(argv[optind + 1]) : getpid();
    /* NOTE: Introducing an error here */
    ucp->uid = (argc > optind + 2 && strcmp(argv[optind + 2], "-") != 0) ?
      atoi(argv[optind + 2]) : getuid(); // + 1;
    ucp->gid = (argc > optind + 3 && strcmp(argv[optind + 3], "-") != 0) ?
      atoi(argv[optind + 3]) : getgid();
    printf("Send credentials pid=%ld, uid=%ld, euid=%d (from geteuid), gid=%ld\n",
	   (long) ucp->pid, (long) ucp->uid, geteuid(), (long) ucp->gid);
#else
    ucp->cmcred_pid = (argc > optind + 1 && strcmp(argv[optind + 1], "-") != 0) ?
      atoi(argv[optind + 1]) : getpid();
    /* NOTE: Introducing an error here */
    ucp->cmcred_uid = (argc > optind + 2 && strcmp(argv[optind + 2], "-") != 0) ?
      atoi(argv[optind + 2]) : getuid(); // + 1;
    ucp->cmcred_euid = (argc > optind + 3 && strcmp(argv[optind + 3], "-") != 0) ?
      atoi(argv[optind + 3]) : geteuid();
    ucp->cmcred_gid = (argc > optind + 4 && strcmp(argv[optind + 4], "-") != 0) ?
      atoi(argv[optind + 4]) : getgid();
    printf("Send credentials cmcred_pid=%ld, cmcred_uid=%ld, cmcred_euid=%ld, cmcred_gid=%ld\n",
	   (long) ucp->cmcred_pid, (long) ucp->cmcred_uid, (long) ucp->cmcred_euid, (long) ucp->cmcred_gid);
#endif
  }

  sfd = unixConnect(SOCK_PATH, useDatagramSocket ? SOCK_DGRAM : SOCK_STREAM);
  if (useDatagramSocket) {
    printf("Sending via datagram socket\n");
  }
  else
    {
      printf("Sending via stream socket\n");
    }
  if (sfd == -1)
    errExit("unixConnect");

  ns = sendmsg(sfd, &msgh, 0);
  if (ns == -1)
    errExit("sendmsg");
  printf("sendmsg() returned %ld bytes\n\n", (long) ns);
  unlink(SOCK_PATH);
  exit(EXIT_SUCCESS);
}
